import itertools
import collections

from . import monosaccharide, constants
from .link import LinkMaskContext
from glypy.composition import Composition, structure_composition
from glypy.utils.multimap import OrderedMultiMap

RingType = constants.RingType
SuperClass = constants.SuperClass
Configuration = constants.Configuration
Stem = constants.Stem
Anomer = constants.Anomer

UnknownPosition = constants.UnknownPosition

Monosaccharide = monosaccharide.Monosaccharide
ReducedEnd = monosaccharide.ReducedEnd
Modification = monosaccharide.Modification
Link = monosaccharide.Link
graph_clone = monosaccharide.graph_clone
traverse = monosaccharide.traverse
toggle = monosaccharide.toggle
SuperClass = constants.SuperClass
modification_compositions = structure_composition.modification_compositions

chain = itertools.chain
Counter = collections.Counter


class CrossRingPair(object):
    @classmethod
    def can_crossring_fragment(cls, link):
        """Test if the given |Link| object's :attr:`child` supports
        cross ring fragments.

        Parameters
        ----------
        link : Link
            |Link| object whose child will be cross ring cleaved

        Returns
        -------
        bool
        """
        child = link.child
        ring_type = child.ring_type
        return ring_type is not RingType.x and ring_type is not RingType.open

    @classmethod
    def from_link(cls, link):
        """Generate an instance of :class:`CrossRingPair` for each method
        in which the :attr:`child` of `link` can be cross ring cleaved.

        Parameters
        ----------
        link : Link
            |Link| object whose child will be cross ring cleaved

        Yields
        ------
        :class:`CrossRingPair` :
            An instance of :class:`CrossRingPair` for the next viable
            set of ring coordinates
        """
        if not cls.can_crossring_fragment(link):
            return []
        return [cls(link.child, c1, c2) for c1, c2 in
                enumerate_cleavage_pairs(link.child)]

    def __init__(self, residue, c1, c2):
        self.residue = residue
        self.id = residue.id
        self.toggler = None
        self.parent = None
        self.child = None
        self.active = False
        self.cleave_1 = c1
        self.cleave_2 = c2

    def break_link(self, **kwargs):
        """Emulate |Link| interface for passing state with a call to
        :func:`crossring_fragments`.

        Call :func:`crossring_fragments` with :attr:`cleave_1` and :attr:`cleave_2` on
        :attr:`residue` in-place.

        Sets :attr:`parent` to the resulting `X` fragment and :attr:`child` to the resulting
        `A` fragment.

        Sets :attr:`toggler` to an instance of :class:`glypy.structure.link.LinkMaskContext`
        and calls its :meth:`mask` method to hide the links attached to :attr:`residue` after
        they have been duplicated on the resulting fragments.

        Returns
        -------
        :class:`CrossRingFragment` parent: The X fragment from :func:`crossring_fragments`
        :class:`CrossRingFragment` child: The A fragment from :func:`crossring_fragments`
        """
        a, x = crossring_fragments(self.residue, self.cleave_1, self.cleave_2, copy=False)
        # Bind the toggler late to pick up any changes to the residue's links after instantiation
        self.toggler = LinkMaskContext(self.residue)
        self.toggler.mask()
        self.parent = x
        self.child = a
        self.parent._pair = self
        self.child._pair = self
        self.active = True
        return self.parent, self.child

    def apply(self):
        """Emulate |Link| interface for passing state with the removal of
        links to the associated :class:`CrossRingFragment` objects.

        Calls :meth:`CrossRingFragment.release` on :attr:`parent` and :attr:`child`.

        Calls :meth:`LinkMaskContext.unmask`.

        .. warning::
            This method cannot find |Link| objects that are already hidden under another
            :class:`LinkMaskContext`. If using multiple masks on adjacent nodes, make sure
            to rerun :meth:`CrossRingPair.release` again after all links are unmasked.

        """
        self.toggler.unmask()
        self.parent.release()
        self.child.release()
        self.active = False
        self.toggler = None

    def release(self):
        """Convenience function to call :meth:`CrossRingFragment.release` on
        :attr:`parent` and :attr:`child`
        """
        self.parent.release()
        self.child.release()
        self.active = False

    def __repr__(self):  # pragma: no cover
        rep = "<CrossRingPair id={} {},{}>".format(self.id, self.cleave_1, self.cleave_2)
        return rep

    def is_attached(self, deep=False):
        return len(self.child.links) + len(self.parent.links) > 0


class CrossRingFragment(Monosaccharide):
    '''
    Describes a fragment formed by cleaving across two bonds of the ring of a
    cyclical monosaccharide. Behaves in all respects like an instance of |Monosaccharide|,
    in addition to the following adjustments.

    Attributes
    ----------
    cleave_1, cleave_2: int
        Ring sites that were cleaved
    kind: str
        Whether the fragment is reducing (X) or non-reducing (A)
    contains: list
        The list of positions on the monosaccharide included in this fragment
    _link_cache: list
        A list of all |Link| objects connecting this fragment to other |Monosaccharide| objects
        regardless of whether they are attached. Convenient for ease of attaching and
        detatching to a |Glycan|
    '''
    def __init__(self, composition, cleave_1, cleave_2, contains, kind,
                 modifications=None, anomer=Anomer.x,
                 stem=(Stem.x,), configuration=(Configuration.x,), id=None,
                 link_cache=None, source=None):
        _base_composition = composition.clone()
        super(CrossRingFragment, self).__init__(
            modifications=modifications, ring_start=None, ring_end=None,
            substituent_links=None, links=None, superclass=SuperClass.x, anomer=anomer,
            stem=stem, configuration=configuration, id=id, composition=composition, fast=True)
        self._source = source
        self._base_composition = _base_composition
        self.kind = kind
        self.cleave_1 = cleave_1
        self.cleave_2 = cleave_2
        self.contains = contains
        self._link_cache = link_cache or []

    def attach(self):
        '''
        Calls :meth:`glypy.structure.link.Link.apply` for each |Link| in :attr:`_link_cache`
        '''
        for link in self._link_cache:
            if not link.is_attached():
                link.apply()

    def release(self):
        '''
        Calls :meth:`glypy.structure.link.Link.break_link` with `refund=True` for each |Link| in :attr:`_link_cache`
        '''
        for link in self.links.values():
            if link.is_attached():
                link.break_link(refund=True)
        for link in self._link_cache:
            if link.is_attached():
                link.break_link(refund=True)

    def clone(self, prop_id=False, fast=True):
        """Clone this instance, excluding :attr:`links`

        Returns
        -------
        CrossRingFragment

        See Also
        --------
        :meth:`glypy.structure.monosaccharide.Monosaccharide.clone`
        """
        modifications = OrderedMultiMap()
        for k, v in self.modifications.items():
            if isinstance(v, ReducedEnd):
                continue
            modifications[k] = v

        residue = CrossRingFragment(
            self._base_composition.clone(), self.cleave_1, self.cleave_2,
            self.contains, self.kind,
            stem=self._stem,
            configuration=self._configuration,
            modifications=modifications,
            anomer=self._anomer,
            id=self.id if prop_id else None)
        for pos, link in self.substituent_links.items():
            sub = link.to(self)
            dup = sub.clone()
            link.clone(residue, dup)

        return residue

    def __repr__(self):  # pragma: no cover
        return "CrossRingFragment({kind}({c1}, {c2}) {contains} {mass})".format(
            kind=self.kind, c1=self.cleave_1, c2=self.cleave_2, contains=self.contains, mass=self.mass())

    def __eq__(self, other):
        """Test for equality between `self` and `other`.

        Parameters
        ----------
        other

        Returns
        -------
        bool
        """
        if other is None:
            return False
        res = super(CrossRingFragment, self).__eq__(other)

        return res

    def __ne__(self, other):
        return not self == other

    def graph_mass(self, average=False, mass_data=None):
        '''
        Calcules the total mass of all connected residues.

        See Also
        --------
        glypy.structure.glycan.Glycan.mass
        '''
        charge = 0
        return sum(
            node.mass(average=average, charge=charge, mass_data=mass_data) for node in traverse(self))

    def open_attachment_sites(self, max_occupancy=0):
        slots = Counter()
        for i in self.contains:
            slots[i] = 0
        unknowns = 0
        null_positions = (UnknownPosition, 'x')
        for pos, obj in chain(self.modifications.items(),
                              self.links.items(),
                              self.substituent_links.items()):
            if obj == Modification.keto:
                continue
            if pos in null_positions:
                unknowns += 1
            else:
                slots[pos] += 1

        open_slots = []

        for i, count in slots.items():
            if count <= max_occupancy:
                open_slots.append(i)

        return open_slots, unknowns


def crossring_fragments(monosaccharide, c1, c2, attach=True, copy=True):
    '''
    Generate cross-ring fragments from `monosaccharide` cutting bonds at `c1` and `c2`. If
    `attach` is |True|, bonds will be attached between the resulting :class:`CrossRingFragment` and
    other |Monosaccharide| objects. Otherwise they will be created but not `apply`'d. If `copy` is |True|,
    then attached subtrees will be cloned, otherwise they will be shared with the original residue.

    Parameters
    ----------
    monosaccharide: Monosaccharide
        Residue to be cleaved
    c1, c2: int
        Sites to be cleaved at
    attach: bool
        Whether or not |Link| objects between the resulting :class:`CrossRingFragment` objects should
        be made active by `apply` or just saved. Defaults to |True|
    copy: bool
        Whether or not to clone subtrees or share the originals. Defaults to |True|

    Returns
    -------
    tuple of CrossRingFragment
    '''
    ring_type = monosaccharide.ring_type
    if ring_type is RingType.x or ring_type is RingType.open:
        raise TypeError("Cannot cleave an open or unknown carbohydrate backbone")

    c1_segment, c1_include, c2_segment, c2_include = cleave_ring(
        monosaccharide, c1, c2)

    c1_fragment = pack_fragment(c1_segment, c1, c2, c1_include, monosaccharide, attach=attach, copy=copy)
    c2_fragment = pack_fragment(c2_segment, c1, c2, c2_include, monosaccharide, attach=attach, copy=copy)

    ring_start = monosaccharide.ring_start
    ring_end = monosaccharide.ring_end

    a_fragment = x_fragment = None
    if ring_start in c1_fragment.contains:
        x_fragment = c1_fragment
        x_fragment.kind = "X"
        a_fragment = c2_fragment
        a_fragment.kind = "A"
    else:
        x_fragment = c2_fragment
        x_fragment.kind = "X"
        a_fragment = c1_fragment
        a_fragment.kind = "A"

    # If the ring_end is cleaved, then the O component of
    # the ring should go with the X ion
    if ring_end - (ring_start - 1) == c2:
        c1_fragment.composition -= {"O": 1}
        c1_fragment._base_composition -= {"O": 1}
        c2_fragment.composition += {"O": 1}
        c2_fragment._base_composition += {"O": 1}

    return a_fragment, x_fragment


def enumerate_cleavage_pairs(residue):
    '''
    Enumerate all positions `residue` can be cross-ring cleaved at
    '''
    ring_size = (residue.ring_end - residue.ring_start) + 2
    for c1, c2 in itertools.combinations(range(0, ring_size), 2):
        if (c2 - c1 > 1) and ((c2 - c1) + 1 != ring_size):
            yield c1, c2


def pack_fragment(fragment_parts, c1, c2, include, residue, attach=True, copy=True):
    '''
    Given a collection of parts from an unrolled and cleaved ring, assemble them
    into a :clas:`CrossRingFragment` and apply any links or modifications observed
    in the parent residue.

    Parameters
    ----------
    fragment_parts: list
        A list of |dict| objects containing carbohydrate backbone compositions, modification and
        link attachments.
    c1, c2: int
        Sites cleaved at
    include: list
        A list of |int| values referring to the carbons included in this fragment
    residue: Monosaccharide
        The parent |Monosaccharide|. Passed to provide access to :attr:`.Monosaccharide.links`
        for constructing subtrees.

    Returns
    -------
    CrossRingFragment
    '''
    fragment_data = {
        "composition": sum((c['backbone'] for c in fragment_parts), {}),
        "modifications": OrderedMultiMap(),
        "substituent_links": OrderedMultiMap(),
        "links": OrderedMultiMap(),
        "contains": include
    }

    for i in fragment_parts:
        ilink = i['links']
        if ilink:
            fragment_data['links'].update(ilink)

        imodifications = i['modifications']
        if imodifications:
            fragment_data['modifications'].update(imodifications)

        isubstituents = i['substituent_links']
        if isubstituents:
            fragment_data['substituent_links'].update(isubstituents)

    fragment_object = CrossRingFragment(
        fragment_data['composition'], c1, c2, fragment_data['contains'],
        '?', fragment_data['modifications'], stem=residue._stem,
        configuration=residue._configuration, id=residue.id, source=residue
    )

    composition_shift = Composition()
    for pos, mod in fragment_data["modifications"].items():
        composition_shift += modification_compositions[mod](pos)
    fragment_object.composition += composition_shift

    for pos, link in fragment_data['substituent_links'].items():
        subst = link[residue]
        link.clone(fragment_object, subst.clone())

    links = []
    # The id values of all neighboring residues and the parent residue
    edges = {node[residue].id for node in residue.links.values()} | {residue.id}
    for pos, link in fragment_data['links'].items():
        if copy:
            # Copy all structures along the current link excluding all neighbors
            # but the current one to prevent cycles.
            subtree = graph_clone(link[residue], visited=edges - {link[residue].id})
        else:
            subtree = link[residue]
        # Attach the fragment to the subtree and save the link object to the link cache

        if link.is_parent(residue):
            links.append(link.clone(fragment_object, subtree, attach=attach))
        else:
            links.append(link.clone(subtree, fragment_object, attach=attach))
    fragment_object._link_cache = links

    return fragment_object


def unroll_ring(residue):
    '''
    Flattens `residue` into a list of |dict| objects containing
    information about the `i`th position on the carbohydrate backbone

    Parameters
    ----------
    residue: Monosaccharide
        The carbohydrate to unroll

    Returns
    -------
    list of dict
    '''
    ring_size = residue.superclass.value
    nodes = [None for i in range(ring_size)]
    for i in range(ring_size):
        segment = {
            "index": i + 1,
            "backbone": Composition({'H': 2, 'C': 1, 'O': 1}),
            "modifications": OrderedMultiMap(),
            "substituent_links": OrderedMultiMap(),
            "links": OrderedMultiMap()
        }
        for mod in (residue.modifications[i + 1]):
            segment['modifications'][i + 1] = mod

        for sub_link in residue.substituent_links[i + 1]:
            segment['substituent_links'][i + 1] = sub_link

        for glyco_link in residue.links[i + 1]:
            segment['links'][i + 1] = glyco_link

        nodes[i] = segment

    return nodes


def cleave_ring(residue, c1, c2):
    '''
    Given a |Monosaccharide| `residue` and two ring cleavage sites, `(c1, c2)`, calculate
    the linear structures produced by cleaving the ring of `residue`, including any non-ring
    carbons with the appropriate fragment. Calls :func:`slice_ring` and :func:`unroll_ring`

    Parameters
    ----------
    residue: Monosaccharide
        The residue to be cleaved
    c1, c2: int
        The cleavage sites

    Returns
    -------
    c1_segment: list of dicts of backbone information
    c1_include: list of ints
    c2_segment: list of dicts of backbone information
    c2_include: list of ints
    '''

    linear = unroll_ring(residue)
    ring_start = residue.ring_start
    ring_end = residue.ring_end

    # Slice the ring region of the linearized carbohydrate
    c1_segment, c1_include, c2_segment, c2_include = slice_ring(
        linear[ring_start - 1: ring_end], c1, c2)

    # If the ring's start is associated with c1, add all positions prior to ring start
    # to c1.
    if ring_start in c1_include and ring_start - 1 not in c1_include and ring_start != 1:
        c1_segment = linear[:ring_start - 1] + c1_segment
        c1_include = list(i + 1 for i in range(ring_start + 1)) + c1_include
    # Alternatively if the ring start is in c2, add the preceding residues to it
    elif ring_start in c2_include and ring_start - 1 not in c2_include and ring_start != 1:
        c2_segment = linear[:ring_start - 1] + c2_segment
        c2_include = list(i + 1 for i in range(ring_start + 1)) + c2_include

    # If the ring's end is associated with c2, add all positions proceeding ring end
    # to c2
    if ring_end in c2_include:
        c2_segment = c2_segment + linear[ring_end:]
        c2_include = c2_include + \
            list(i + 1 for i in range(ring_end, len(linear)))
    # else they must be in c1
    else:
        c1_segment = c1_segment + linear[ring_end:]
        c1_include = c1_include + \
            list(i + 1 for i in range(ring_end, len(linear)))

    return c1_segment, c1_include, c2_segment, c2_include


def slice_ring(linear, c1, c2):
    '''
    Given a linear carbohydrate and two cut sites, extract the two regions
    produced by cutting the sequence like a ring buffer.
    '''
    # The first fragment is defined as the region between the first and
    # second cut.
    c1_segment = linear[c1:c2]
    c1_include = [c['index'] for c in c1_segment]
    # Add the positions after the second cut up to the end of the ring
    # and the positions from the start of the ring to the first cut to the
    # second fragment.
    c2_segment = [linear[i]
                  for i in list(range(c2, len(linear))) + list(range(c1))
                  ]
    c2_include = [c['index'] for c in c2_segment]
    return c1_segment, c1_include, c2_segment, c2_include



import pkg_resources
import hjson
import re

from glypy.utils import StringIO, identity, uid
from glypy.utils.lazy import ProxyObject
from glypy.io import glycoct
from glypy.structure.glycan import NamedGlycan


class StructureIndex(dict):
    def __init__(self, stream, key_transform=identity, value_transform=identity):
        self.update(hjson.load(stream))
        for k, v in self.items():
            self[key_transform(k)] = value_transform(glycoct.loads(v))
        self.key_transform = key_transform

    def __getitem__(self, key):
        x = dict.__getitem__(self, key)
        # ret = deepcopy(x)
        ret = x.clone()
        ret.id = uid()
        return ret

    def __getattr__(self, name):
        try:
            res = object.__getattr__(self, name)
            return res
        except AttributeError:
            return self[name]

    def __dir__(self):
        return list(self.__dict__) + list(k for k in self
                                          if (not k[0].isdigit()) and (not re.search(r"[:,\-\s\(\)\[\]]", k)))

    def __repr__(self):
        return "<{}>".format(self.__class__.__name__)

    def __str__(self):  # pragma: no cover
        rep = StringIO()
        fmt = "{0}\n{1}\n{2}\n\n"
        for k, v in self.items():
            rep.write(fmt.format(k, v.to_glycoct(), v.mass(True)))
        return rep.getvalue()


class MonosaccharideIndex(StructureIndex):
    def __init__(self, stream=None, key_transform=identity, value_transform=lambda x: x.root):
        if stream is None:
            stream = pkg_resources.resource_stream(__name__, "data/monosaccharides.hjson")
        with stream:
            super(MonosaccharideIndex, self).__init__(stream, key_transform, value_transform)


monosaccharides = (MonosaccharideIndex)()


class MonosaccharideResidueIndex(MonosaccharideIndex):
    def __init__(self, stream=None, **kwargs):
        from glypy.structure.glycan_composition import MonosaccharideResidue

        key_transform = identity

        def value_transform(x):
            return MonosaccharideResidue.from_monosaccharide(x.root)

        if stream is None:
            stream = pkg_resources.resource_stream(__name__, "data/monosaccharides.hjson")
        super(MonosaccharideIndex, self).__init__(stream, key_transform, value_transform)


monosaccharide_residues = ProxyObject(MonosaccharideResidueIndex)


class GlycanIndex(StructureIndex):
    def __init__(self, stream=None, key_transform=identity, value_transform=identity):
        if stream is None:
            stream = pkg_resources.resource_stream(__name__, "data/glycans.hjson")
        super(GlycanIndex, self).__init__(stream, key_transform, value_transform)


glycans = ProxyObject(GlycanIndex)


class MotifIndex(StructureIndex):
    def __init__(self, stream=None, key_transform=identity, value_transform=identity):
        if stream is None:
            stream = pkg_resources.resource_stream(__name__, "data/motifs.hjson")
        with stream:
            data = hjson.load(stream)
        motif_classes = set()
        motif_categories = set()
        for motif in data:
            name = motif['name']
            motif_class = motif['class']
            motif_category = motif['category']
            motif_structure = NamedGlycan(name=name, root=glycoct.loads(motif['glycoct']).root, index_method=None)
            motif_structure.motif_name = name
            motif_structure.motif_class = motif_class
            motif_structure.motif_category = motif_category
            motif_structure.is_core_motif = motif["core_motif"]
            self[name] = motif_structure
            motif_classes.add(motif_class)
            motif_categories.add(motif_category)
        self._category_map = {}
        self._class_map = {}
        self.motif_classes = motif_classes
        self.motif_categories = motif_categories

    def motif_category(self, name):
        if name in self._category_map:
            return self._category_map[name]

        mapping = {}
        for k, v in self.items():
            if v.motif_category == name:
                mapping[k] = v
        if len(mapping) > 0:
            self._category_map[name] = mapping
        return mapping

    def motif_class(self, name):
        if name in self._class_map:
            return self._class_map[name]
        mapping = {}
        for k, v in self.items():
            if v.motif_class == name:
                mapping[k] = v
        if len(mapping) > 0:
            self._class_map[name] = mapping
        return mapping


motifs = ProxyObject(MotifIndex)



'''
Defined by `Extended Stereocode <http://monosaccharidedb.org/notation.action?topic=stereocode#ext_stereocode>`_

Symbol  Description
h   "head or tail group", CH2OH group at a terminal position
d   DEOXY core modification at non-terminal position
m   DEOXY core modification at terminal position ("methyl" group)
a   ACID core modification
o   aldehyde group
k   KETO core modification at non-terminal position
e   EN + deoxy core modifications
n   EN core modification without DEOXY core modification
E   EN core modification with unknown deoxygenation status
y   YN core modification at non-terminal position
s   SP2 core modification
t   SP core modification (always at terminal position)
1   "L-Configuration" carbon atom
2   "D-Configuration" carbon atom
x   unknown configuration (D or L) carbon atom
'''

import warnings
from six import string_types as basestring
from six.moves import range

from glypy.utils.enum import EnumValue

from .constants import Stereocoding, Modification, Anomer, Configuration, Stem, UnknownPosition, NoPosition


def sdecode(code):
    return [Stereocoding[c] for c in code]


def invert(code):
    out = []
    L = 1
    D = 2
    for c in code:
        if c == L:
            out.append(D)
        elif c == D:
            out.append(L)
        else:
            out.append(c)
    return sdecode(out)


reference_stereomap = {
    ('alpha', 'l', (('man',), 'hex')): '122110',
    ('alpha', 'd', (('man',), 'hex')): '211220',
    ('beta', 'd', (('man',), 'hex')): '111220',
    ('beta', 'l', (('man',), 'hex')): '222110',
    ('uncyclized', 'd', (('gal',), 'hex')): '021120',
    ('alpha', 'l', (('glc',), 'hex')): '112110',
    ('beta', 'l', (('ara',), 'pen')): '22110',
    ('beta', 'l', (('alt',), 'hex')): '221110',
    ('alpha', 'l', (('gro', 'man'), 'hep')): '2112210',
    ('beta', 'd', (('gro', 'gal'), 'non')): '010211220',
    ('alpha', 'd', (('gal',), 'hex')): '221120',
    ('beta', 'd', (('gul',), 'hex')): '122120',
    ('alpha', 'd', (('ido',), 'hex')): '212120',
    ('alpha', 'd', (('rib',), 'hex')): '220220',
    ('beta', 'd', (('all',), 'hex')): '122220',
    ('alpha', 'd', (('all',), 'hex')): '222220',
    ('beta', 'd', (('xyl',), 'hex')): '121020',
    ('beta', 'd', (('gal',), 'hex')): '121120',
    ('alpha', 'd', (('gro', 'gal'), 'non')): '020211220',
    ('alpha', 'd', (('gro', 'alt'), 'hep')): '2122220',
    ('beta', 'd', (('thr',), 'hex')): '100120',
    ('alpha', 'l', (('ara',), 'pen')): '12110',
    ('alpha', 'l', (('lyx',), 'hex')): '102210',
    ('alpha', 'd', (('ara',), 'hex')): '021220',
    ('beta', 'l', (('glc',), 'hex')): '212110',
    ('beta', 'l', (('ara',), 'hex')): '202110',
    ('uncyclized', 'd', (('ery',), 'pen')): '00220',
    ('beta', 'd', (('rib',), 'hex')): '102220',
    ('uncyclized', 'd', (('glc',), 'hex')): '021220',
    ('alpha', 'l', (('tal',), 'hex')): '122210',
    ('alpha', 'd', (('glc',), 'hex')): '221220',
    ('alpha', 'd', (('gro', 'man'), 'hep')): '2112220',
    ('alpha', 'd', (('rib',), 'pen')): '22220',
    ('beta', 'd', (('lyx',), 'hex')): '101120',
    ('alpha', 'd', (('man',), 'oct')): '02011220',
    ('beta', 'l', (('gul',), 'hex')): '211210',
    ('beta', 'l', (('rib',), 'hex')): '201110',
    ('alpha', 'd', (('xyl',), 'pen')): '22120',
    ('alpha', 'l', (('gal',), 'hex')): '112210',
    ('alpha', 'l', (('ido',), 'hex')): '121210',
    ('beta', 'l', (('ido',), 'hex')): '221210',
    ('beta', 'l', (('gal',), 'hex')): '212210',
    ('uncyclized', 'l', (('ara',), 'pen')): '02110',
    ('alpha', 'l', (('rib',), 'hex')): '101110',
    ('uncyclized', 'd', (('thr',), 'pen')): '00120',
    ('beta', 'l', (('thr',), 'hex')): '221000',
    ('beta', 'd', (('xyl',), 'pen')): '12120',
    ('alpha', 'l', (('gul',), 'hex')): '111210',
    ('beta', 'd', (('glc',), 'hex')): '121220',
    ('beta', 'l', (('tal',), 'hex')): '222210',
    ('alpha', 'l', (('ara',), 'hex')): '102110',
    ('beta', 'd', (('ara',), 'hex')): '011220',
    ('beta', 'd', (('man',), 'oct')): '01011220',
    ('alpha', 'd', (('man',), 'oct')): '02011220',
}

reference_stereomap = {k: sdecode(v) for k, v in reference_stereomap.items()}


def get_stereocode_key(monosaccharide, anomer=None, configuration=None):
    if anomer is None:
        anomer = monosaccharide.anomer
    if configuration is None:
        configuration = monosaccharide.configuration[0]
    key = (anomer, configuration,
           (monosaccharide.stem, monosaccharide.superclass))
    return key


def _update_stereocode_basic(code, monosaccharide):
    for i in range(len(code)):
        modifications = monosaccharide.modifications[i + 1]
        for mod in modifications:
            if mod.name == Modification.d:
                code[i] = 0
            elif mod.name == Modification.a.name:
                code[i] = 0
    return code


def _update_stereocode_extended(code, monosaccharide):
    null_position = (UnknownPosition, NoPosition)
    if monosaccharide.ring_start not in null_position and monosaccharide.ring_end not in null_position:
        ring_range = range(monosaccharide.ring_start, monosaccharide.ring_end + 1)
    else:
        ring_range = []
    ring_dimensions = range(1, monosaccharide.superclass.value + 1)
    termini = set(ring_dimensions) - set(ring_range)
    terminal = max(termini) - 1
    code[terminal] = Stereocoding.h
    for i in range(len(code)):
        modifications = monosaccharide.modifications[i + 1]
        for mod in modifications:
            if mod.name == Modification.d.name:
                if (i + 1) in ring_range:
                    code[i] = Stereocoding.d
                else:
                    code[i] = Stereocoding.m
            elif mod.name == Modification.a:
                code[i] = Stereocoding.a
    return code


class Stereocode(object):
    def __init__(self, code):
        if isinstance(code, basestring):
            code = sdecode(code)
        self.code = list(code)

    def __getitem__(self, i):
        return self.code[i]

    def __setitem__(self, i, v):
        if not isinstance(v, EnumValue):
            self.code[i] = Stereocoding[str(v)]
        else:
            self.code[i] = Stereocoding[v.name]

    def __iter__(self):
        return iter(self.code)

    def __len__(self):
        return len(self.code)

    def basic_encoding(self):
        encoding = {
            Stereocoding.x: 'x',
            Stereocoding.h: '0',
            Stereocoding.a: '0',
            Stereocoding.d: '0'
        }

        return ''.join(encoding.get(c, c.value) for c in self.code)

    def extended_encoding(self):
        return ''.join(c.value if c.value is not None else 'x' for c in self.code)

    def __str__(self):
        return self.extended_encoding()

    def __repr__(self):
        return "Stereocode(%r)" % (self.extended_encoding(),)


def _resolve_unknown_anomer(monosaccharide):
    configuration = monosaccharide.configuration[0]
    has_known_configuration = (configuration != Configuration.x)
    if not has_known_configuration:
        configuration = Configuration.d
    a_key = get_stereocode_key(monosaccharide, Anomer.a, configuration)
    b_key = get_stereocode_key(monosaccharide, Anomer.b, configuration)
    a_code = reference_stereomap[a_key]
    b_code = reference_stereomap[b_key]
    fill_value = Stereocoding[None]
    if monosaccharide.anomer == Anomer.uncyclized:
        fill_value = Stereocoding.h
    substituted_code = [ai if ai == bi else fill_value for ai, bi in zip(a_code, b_code)]
    if not has_known_configuration:
        translate_unknown_configuration = {
            Stereocoding.L: Stereocoding.LD,
            Stereocoding.D: Stereocoding.DL,

        }
        substituted_code = [
            translate_unknown_configuration.get(i, i) for i in substituted_code
        ]
    return substituted_code


def stereocode(monosaccharide):
    key = get_stereocode_key(monosaccharide)
    try:
        if monosaccharide.stem[0] == Stem.x:
            base_code = Stereocode(Stereocoding[None] for i in range(monosaccharide.superclass.value))
        else:
            try:
                base_code = Stereocode(reference_stereomap[key])
            except KeyError:
                base_code = _resolve_unknown_anomer(monosaccharide)
    except KeyError:
        warnings.warn("Could not locate a stereocode template for %r based upon key %s" % (
            monosaccharide, key), stacklevel=3)
        base_code = Stereocode(Stereocoding[None] for i in range(monosaccharide.superclass.value))
    _update_stereocode_extended(base_code, monosaccharide)
    return base_code



import os
import unittest
import matplotlib
matplotlib.use("Agg")
from matplotlib import pyplot as plt

from glypy import monosaccharides, glycans
from glypy.io import glycoct
from glypy import plot


def outplot_dir():
    wd = os.getcwd()
    if "test_data" in os.listdir(wd):
        return "test_data" + os.sep
    elif "test_data" in os.listdir(wd + os.sep + '..'):
        return wd + os.sep + '..' + os.sep + "test_data"
    else:
        os.mkdir("test_data")
        return "test_data" + os.sep

plot_dir = outplot_dir()


glycan_structure = glycoct.loads('''
RES
1b:x-dglc-HEX-1:5
2s:n-acetyl
3b:b-dglc-HEX-1:5
4s:n-acetyl
5b:b-dman-HEX-1:5
6b:a-dman-HEX-1:5
7b:b-dglc-HEX-1:5
8s:n-acetyl
9b:a-lgal-HEX-1:5|6:d
10b:b-dgal-HEX-1:5
11b:a-dgro-dgal-NON-2:6|1:a|2:keto|3:d
12s:n-glycolyl
13b:b-dglc-HEX-1:5
14s:n-acetyl
15b:b-dgal-HEX-1:5
16s:n-acetyl
17b:b-dglc-HEX-1:5
18s:n-acetyl
19b:a-dman-HEX-1:5
20b:b-dglc-HEX-1:5
21s:n-acetyl
22b:a-lgal-HEX-1:5|6:d
23b:b-dgal-HEX-1:5
24b:a-dgro-dgal-NON-2:6|1:a|2:keto|3:d
25s:n-glycolyl
26b:b-dglc-HEX-1:5
27s:n-acetyl
28b:a-lgal-HEX-1:5|6:d
29b:b-dgal-HEX-1:5
30b:a-dgro-dgal-NON-2:6|1:a|2:keto|3:d
31s:n-acetyl
32b:a-lgal-HEX-1:5|6:d
LIN
1:1d(2+1)2n
2:1o(4+1)3d
3:3d(2+1)4n
4:3o(4+1)5d
5:5o(3+1)6d
6:6o(2+1)7d
7:7d(2+1)8n
8:7o(3+1)9d
9:7o(4+1)10d
10:10o(3+2)11d
11:11d(5+1)12n
12:6o(4+1)13d
13:13d(2+1)14n
14:13o(4+1)15d
15:15d(2+1)16n
16:5o(4+1)17d
17:17d(2+1)18n
18:5o(6+1)19d
19:19o(2+1)20d
20:20d(2+1)21n
21:20o(3+1)22d
22:20o(4+1)23d
23:23o(3+2)24d
24:24d(5+1)25n
25:19o(6+1)26d
26:26d(2+1)27n
27:26o(3+1)28d
28:26o(4+1)29d
29:29o(3+2)30d
30:30d(5+1)31n
31:1o(6+1)32d
''')


class PlotTests(unittest.TestCase):
    def test_plot_monosaccharide(self):
        for name, residue in monosaccharides.items():
            plot.plot(residue)
            plt.clf()
            plt.close('all')

    def test_draw_tree(self):
        dt = plot.DrawTree(glycan_structure)
        self.assertEqual(dt.get(1).tree, glycan_structure.get(1))
        parent, child = dt.get_link_pair(1)
        parent = parent.tree
        child = child.tree
        self.assertEqual(parent, glycan_structure.get(1))

    def test_plot_iupac(self):
        plot.plot(glycan_structure, symbol_nomenclature='iupac')

    def test_plot_glycans(self):
        for name, structure in glycans.items():
            for layout in ['balanced', 'topological']:
                for symbol_nomenclature in ['cfg']:  # ['cfg', 'iupac']:
                    plot.plot(
                        structure, label=True, orientation='h',
                        symbol_nomenclature=symbol_nomenclature, layout=layout)
                    path = "%s/%s_%s_%s.png" % (plot_dir, name, layout, symbol_nomenclature)
                    plt.savefig(path)
                    plt.clf()
                    plt.close('all')

        plt.figure()
        plot.enumerate_tree(*plot.plot(glycan_structure))
        plt.savefig(plot_dir + os.sep + "enum_tree" + '.png')


if __name__ == '__main__':
    unittest.main()



from . import cfg_symbols, snfg_symbols
from .draw_tree import plot, DrawTreeNode, enumerate_tree, DrawTree
from .cfg_symbols import CFGNomenclature
from .snfg_symbols import SNFGNomenclature
from .symbolic_nomenclature import SymbolicNomenclatureBase
from .geometry import TreeLayoutBase
from .buchheim import BalancedTreeLayout
from .topological_layout import TopologicalTreeLayout
from . import common


__all__ = [
    "plot", "DrawTreeNode", "enumerate_tree",
    "cfg_symbols", "snfg_symbols", 'CFGNomenclature',
    'SNFGNomenclature', 'SymbolicNomenclatureBase',
    'TreeLayoutBase', 'BalancedTreeLayout',
    'TopologicalTreeLayout', 'common'
]



import unittest


import glypy
from glypy.composition import composition_transform
from glypy.tests import common
from glypy.io import iupac, glycoct

monosaccharides = common.monosaccharides


class IUPACTests(unittest.TestCase):
    def test_monosaccharide_to_iupac(self):
        iupac_residue = iupac.to_iupac(monosaccharides.Fucose)
        reference = "a-L-Fucp"
        self.assertEqual(iupac_residue, reference)

    def test_hexa(self):
        text = "a-D-HexpA"
        result = iupac.from_iupac(text)
        self.assertAlmostEqual(result.mass(), 194.042, 2)

    def test_autoname_substituent(self):
        h = monosaccharides.Hex
        h.add_substituent("phosphate")
        iupac.substituents_map_to.pop("phosphate")
        t = iupac.to_iupac(h)
        r = iupac.from_iupac(t)
        self.assertEqual(h, r)

    def test_special_cases(self):
        for text, mass in [('?-D-Neup', 267.095), ('?-D-Neup5Ac', 309.105), ('a-D-Neup5NGc', 325.100)]:
            self.assertAlmostEqual(iupac.from_iupac(
                iupac.to_iupac(iupac.from_iupac(text))).mass(), mass, 2)

    def test_ring_type(self):
        for text, start, stop in [("a-D-Hexp", 1, 5), ("a-D-Hexf", 1, 4), ("a-D-Hexo", 0, 0)]:
            mono = iupac.from_iupac(text)
            self.assertEqual(mono.ring_start, start)
            self.assertEqual(mono.ring_end, stop)

    def test_alternate_superclass(self):
        text = "a-D-2-deoxy-araHex"
        mono = iupac.from_iupac(text)
        self.assertEqual(mono.stem[0], 'ara')

    def test_monosaccharide_from_iupac(self):
        text = "a-L-Fucp"
        reference = monosaccharides.Fucose
        result = iupac.from_iupac(text)
        self.assertEqual(result, reference)

    def test_glycan_to_iupac(self):
        reference = 'a-L-Fucp-(1-6)-[a-D-Neup5Ac-(2-3)-b-D-Galp-(1-4)-[a-L-Fucp-(1-3)]b-D-Glcp2NAc-(1-6)-[a-D-Neup5Gc-(2-3)-b-D-Galp-(1-4)-[a-L-Fucp-(1-3)]b-D-Glcp2NAc-(1-2)]a-D-Manp-(1-6)-[b-D-Glcp2NAc-(1-4)][b-D-Galp2NAc-(1-4)-b-D-Glcp2NAc-(1-4)-[a-D-Neup5Gc-(2-3)-b-D-Galp-(1-4)-[a-L-Fucp-(1-3)]b-D-Glcp2NAc-(1-2)]a-D-Manp-(1-3)]b-D-Manp-(1-4)-b-D-Glcp2NAc-(1-4)]?-D-Glcp2NAc'
        structure = common.load("complex_glycan")
        self.assertEqual(iupac.to_iupac(structure), reference)

    def test_glycan_from_iupac(self):
        text = 'a-L-Fucp-(1-6)-[a-D-Neup5Ac-(2-3)-b-D-Galp-(1-4)-[a-L-Fucp-(1-3)]b-D-Glcp2NAc-(1-6)-[a-D-Neup5Gc-(2-3)-b-D-Galp-(1-4)-[a-L-Fucp-(1-3)]b-D-Glcp2NAc-(1-2)]a-D-Manp-(1-6)-[b-D-Glcp2NAc-(1-4)][b-D-Galp2NAc-(1-4)-b-D-Glcp2NAc-(1-4)-[a-D-Neup5Gc-(2-3)-b-D-Galp-(1-4)-[a-L-Fucp-(1-3)]b-D-Glcp2NAc-(1-2)]a-D-Manp-(1-3)]b-D-Manp-(1-4)-b-D-Glcp2NAc-(1-4)]?-D-Glcp2NAc'
        reference = common.load("complex_glycan")
        self.assertEqual(glycoct.canonicalize(iupac.from_iupac(text)), glycoct.canonicalize(reference))

    def test_glycan_from_iupac_file_like(self):
        text = '''>test
        a-L-Fucp-(1-6)-[a-D-Neup5Ac-(2-3)-b-D-Galp-(1-4)-[a-L-Fucp-(1-3)]b-D-Glcp2NAc-(1-6)-[a-D-Neup5Gc-(2-3)-b-D-Galp-(1-4)-[a-L-Fucp-(1
        -3)]b-D-Glcp2NAc-(1-2)]a-D-Manp-(1-6)-[b-D-Glcp2NAc-(1-4)][b-D-Galp2NAc-(1-4)-b-D-Glcp2NAc-(1-4)-[a-D-Neup5Gc-(2-3)-b-D-Galp-(1-4)-[a-L-Fucp
        -(1-3)]b-D-Glcp2NAc-(1-2)]a-D-Manp-(1-3)]b-D-Manp-(1-4)-b-D-Glcp2NAc-(1-4)]?-D-Glcp2NAc'''
        reader = iupac.IUPACParser.loads(text, 'fasta')
        structure = next(reader)

        text = 'a-L-Fucp-(1-6)-[a-D-Neup5Ac-(2-3)-b-D-Galp-(1-4)-[a-L-Fucp-(1-3)]b-D-Glcp2NAc-(1-6)-[a-D-Neup5Gc-(2-3)-b-D-Galp-(1-4)-[a-L-Fucp-(1-3)]b-D-Glcp2NAc-(1-2)]a-D-Manp-(1-6)-[b-D-Glcp2NAc-(1-4)][b-D-Galp2NAc-(1-4)-b-D-Glcp2NAc-(1-4)-[a-D-Neup5Gc-(2-3)-b-D-Galp-(1-4)-[a-L-Fucp-(1-3)]b-D-Glcp2NAc-(1-2)]a-D-Manp-(1-3)]b-D-Manp-(1-4)-b-D-Glcp2NAc-(1-4)]?-D-Glcp2NAc'
        equiv = next(iupac.IUPACParser.loads(text, 'line'))
        self.assertEqual(equiv, structure)


class DerivatizationAwareIUPACTests(unittest.TestCase):
    def test_monosaccharide_parse(self):
        text = '?-?-Hexp2NAc^Me'
        parser = iupac.DerivatizationAwareMonosaccharideDeserializer()
        obj, _ = parser(text)
        ref = composition_transform.derivatize(glypy.monosaccharides.HexNAc, 'methyl')
        self.assertEqual(obj, ref)

    def test_monosaccharide_serialize(self):
        obj = composition_transform.derivatize(glypy.monosaccharides.HexNAc, 'methyl')
        ref = '?-?-Hexp2NAc^Me'
        serializer = iupac.DerivatizationAwareMonosaccharideSerializer()
        text = serializer(obj)
        self.assertEqual(text, ref)

    def test_derivatized_glycan_parse(self):
        ref = composition_transform.derivatize(
            glypy.motifs["N-Glycan complex 1"], "methyl")
        serializer = iupac.GlycanSerializer(iupac.DerivatizationAwareMonosaccharideSerializer())
        text = serializer(ref)
        deserializer = iupac.GlycanDeserializer(iupac.DerivatizationAwareMonosaccharideDeserializer())
        obj = deserializer(text)
        self.assertEqual(obj, ref)

    def test_compatible_with_non_derivatized(self):
        serializer = iupac.GlycanSerializer(iupac.DerivatizationAwareMonosaccharideSerializer())
        deserializer = iupac.GlycanDeserializer(iupac.DerivatizationAwareMonosaccharideDeserializer())
        self.assertEqual(
            deserializer(
                serializer(
                    glypy.motifs["N-Glycan complex 1"])),
            glypy.motifs["N-Glycan complex 1"])


if __name__ == '__main__':
    unittest.main()



import warnings

from six import string_types as basestring

import glypy

from glypy.io import iupac

from glypy.structure.base import MoleculeBase
from glypy.algorithms import subtree_search
from glypy.algorithms.similarity import commutative_similarity

from glypy.utils import root as proot

from .ec import EnzymeInformation, EnzymeDatabase


def rejecting(*args):
    def checker(structure):
        for subtree in args:
            if subtree_search.subtree_of(subtree, structure, exact=True):
                return False
        return True
    return checker


def reject_on_path(*args):
    def checker(structure, selected_node):
        for subtree in args:
            path = {
                v.id: k for k, v in subtree_search.walk_with(structure, subtree)
            }
            if selected_node.id in path:
                return False
        return True
    return checker


def reject_on_parent(*args):
    def checker(structure, selected_node):
        for t in args:
            for _, parent in selected_node.parents():
                if parent == t:
                    return False
        return True
    return checker


class Glycoenzyme(object):

    def __init__(self, parent_position, child_position, parent, child, terminal=True,
                 identifying_information=None,
                 comparator=None,
                 validator=None):
        if comparator is None:
            comparator = commutative_similarity
        if validator is None:
            def validator(structure):  # pylint: disable=function-redefined
                return True
        if isinstance(identifying_information, basestring):
            identifying_information = EnzymeInformation(
                identifying_information)

        self._parent_position = ()
        self._child_position = ()
        self._parents = ()
        self._child = None

        self.parent_position = parent_position
        self.child_position = child_position
        self.parents = parent
        self.child = child
        self.terminal = terminal
        self.comparator = comparator
        self.validators = self._conform_validator(validator)
        self.identifying_information = identifying_information

    @property
    def parent_position(self):
        return self._parent_position

    @parent_position.setter
    def parent_position(self, value):
        self._parent_position = self._conform_position(value)

    @property
    def child_position(self):
        return self._child_position

    @child_position.setter
    def child_position(self, value):
        self._child_position = self._conform_position(value)

    @property
    def parents(self):
        return self._parent

    @parents.setter
    def parents(self, value):
        self._parent = self._conform_molecule(value)

    @property
    def child(self):
        return self._child

    @child.setter
    def child(self, value):
        self._child = value

    def _conform_validator(self, fn):
        try:
            iter(fn)
        except TypeError:
            fn = (fn,)
        return fn

    def validate_structure(self, structure):
        for fn in self.validators:
            if not fn(structure):
                return False
        return True

    def _conform_position(self, value):
        if value is None:
            return None
        try:
            return tuple(value)
        except TypeError:
            return (value,)

    def _conform_molecule(self, value):
        if value is None:
            return value
        if isinstance(value, MoleculeBase):
            value = (value, )
        else:
            value = tuple(value)
        return value

    def _traverse(self, structure):
        raise NotImplementedError()

    def traverse(self, structure):
        if self.validate_structure(structure):
            return list(self._traverse(structure))
        else:
            return []


class Transferase(Glycoenzyme):

    def __init__(self, parent_position, child_position, parent, child, terminal=True,
                 identifying_information=None, comparator=None, validator=None,
                 parent_node_id=None, site_validator=None, exact=True):
        super(Transferase, self).__init__(
            parent_position, child_position, parent, child, terminal,
            identifying_information, comparator, validator)
        if site_validator is None:
            def site_validator(structure, selected_node):  # pylint: disable=function-redefined
                return True
        if parent_node_id is None:
            parent_node_id = proot(self.parent).id
        self.parent_node_id = parent_node_id
        self.site_validators = self._conform_validator(site_validator)
        self.exact = exact

    def _traverse(self, structure):
        if self.parents is not None:
            for parent in self.parents:
                for node in subtree_search.find_matching_subtree_roots(parent, structure, exact=True):
                    node = self._get_paired_node(node, parent)
                    for parent_position in self.parent_position:
                        if not node.is_occupied(parent_position) and (
                           (len(node.children()) == 0 and self.terminal) or (
                                not self.terminal)) and self.validate_site(structure, node):
                            yield node

    def validate_site(self, structure, node):
        for fn in self.site_validators:
            if not fn(structure, node):
                return False
        return True

    def _get_paired_node(self, node, parent):
        return {k.id: v for k, v in subtree_search.walk_with(node, parent)
                }.get(self.parent_node_id)

    def make_bond(self, node, parent_position=None, child_position=None):
        raise NotImplementedError()

    def apply(self, structure, parent_position=None, child_position=None):
        for node in self.traverse(structure):
            new_structure = structure.clone()
            node = new_structure.get(node.id)
            self.make_bond(node, parent_position, child_position)
            # reset ids, rebuild the index, and standardize traversal order
            new_structure.reindex()
            new_structure.canonicalize()
            yield new_structure

    def __call__(self, structure, parent_position=None, child_position=None):
        return self.apply(
            structure, parent_position=parent_position, child_position=child_position)


class Glycosyltransferase(Transferase):

    def make_bond(self, node, parent_position=None, child_position=None):
        new_monosaccharide = self.child.clone()
        node.add_monosaccharide(
            new_monosaccharide, position=self.parent_position[0],
            child_position=self.child_position[0])


class Substituentransferase(Glycosyltransferase):

    def make_bond(self, node, parent_position=None, child_position=None):
        new_substituent = self.child.clone()
        node.add_substituent(
            new_substituent, position=self.parent_position[0],
            child_position=self.child_position[0])


class Glycosylase(Glycoenzyme):

    @property
    def child(self):
        return self._child

    @child.setter
    def child(self, value):
        self._child = self._conform_molecule(value)

    def _test(self, link):
        return (
            (self.parents is None or any(self.comparator(link.parent, parent) for parent in self.parents)),
            (self.child is None or any(self.comparator(link.child, child) for child in self.child)),
            (self.parent_position is None or link.parent_position in self.parent_position),
            (self.child_position is None or link.child_position in self.child_position)
            (len(link.child.children()) == 0 and self.terminal) or (not self.terminal)
        )

    def _traverse(self, structure):
        for p, link in structure.iterlinks():
            if link.is_ambiguous():
                warnings.warn(
                    "Glycosylase do not support ambiguous linkages at this time.")
            else:
                if (self.parents is None or any(self.comparator(link.parent, parent) for parent in self.parents)) and\
                   (self.child is None or any(self.comparator(link.child, child) for child in self.child)) and\
                   (self.parent_position is None or link.parent_position in self.parent_position) and\
                   (self.child_position is None or link.child_position in self.child_position) and\
                   ((len(link.child.children()) == 0 and self.terminal) or (not self.terminal)):
                    yield link

    def digest(self, link, refund=False):
        link.break_link(refund=refund)

    def apply(self, structure, refund=False):
        for link in self.traverse(structure):
            new_structure = structure.clone()
            link = new_structure.get_link(link.id)
            self.digest(link, refund)
            parent, child = structure.__class__(
                link.parent, index_method=None).reroot(index_method='dfs'), structure.__class__(
                link.child, index_method='dfs')
            yield parent.canonicalize(), child.canonicalize()

    def __call__(self, structure, refund=False):
        return self.apply(structure, refund=refund)


def make_n_glycan_pathway():
    enzdb = EnzymeDatabase._from_static()

    bisecting_glcnac = iupac.loads(
        "b-D-Glcp2NAc-(1-4)-b-D-Manp-(1-4)-b-D-Glcp2NAc-(1-4)-b-D-Glcp2NAc")
    galactose = iupac.loads("b-D-Galp")

    parent = iupac.loads(
        "a-D-Manp-(1-3)-b-D-Manp-(1-4)-b-D-Glcp2NAc-(1-4)-b-D-Glcp2NAc")
    child = iupac.loads("b-D-Glcp2NAc")
    gntI = Glycosyltransferase(2, 1, parent, child, terminal=1,
                               identifying_information=enzdb[2, 4, 1, 101], parent_node_id=6)

    parent = iupac.loads(
        "a-D-Manp-(1-6)-[b-D-Glcp2NAc-(1-2)-a-D-Manp-(1-3)]b-D-Manp-(1-4)-b-D-Glcp2NAc-(1-4)-b-D-Glcp2NAc")
    child = iupac.loads("b-D-Glcp2NAc")
    gntII = Glycosyltransferase(2, 1, parent, child, terminal=True,
                                identifying_information=enzdb[2, 4, 1, 143],
                                parent_node_id=6,
                                validator=rejecting(bisecting_glcnac, galactose))

    parent = iupac.loads(
        "beta-D-Glcp2NAc-(1->2)-alpha-D-Manp-(1->3)-"
        "[alpha-D-Manp-(1->6)]"
        "-beta-D-Manp-(1->4)-beta-D-Glcp2NAc-(1->4)-beta-D-Glcp2NAc")
    child = iupac.loads("b-D-Glcp2NAc")
    gntIII = Glycosyltransferase(4, 1, parent, child, terminal=False, parent_node_id=5,
                                 identifying_information=enzdb[2, 4, 1, 144],
                                 validator=rejecting(bisecting_glcnac))

    parent = iupac.loads(
        "b-D-Glcp2NAc-(1-2)-a-D-Manp-(1-3)-b-D-Manp-(1-4)-b-D-Glcp2NAc-(1-4)-b-D-Glcp2NAc")
    child = iupac.loads("b-D-Glcp2NAc")
    gntIV = Glycosyltransferase(4, 1, parent, child, terminal=False, parent_node_id=6,
                                identifying_information=enzdb[2, 4, 1, 145],
                                validator=rejecting(bisecting_glcnac, galactose))

    parent = iupac.loads(
        "b-D-Glcp2NAc-(1-2)-a-D-Manp-(1-6)-b-D-Manp-(1-4)-b-D-Glcp2NAc-(1-4)-b-D-Glcp2NAc")
    child = iupac.loads("b-D-Glcp2NAc")
    gntV = Glycosyltransferase(6, 1, parent, child, terminal=False, parent_node_id=6,
                               identifying_information=enzdb[2, 4, 1, 155],
                               validator=rejecting(bisecting_glcnac, galactose))

    parent = iupac.loads(
        "b-D-Glcp2NAc-(1-6)-[b-D-Glcp2NAc-(1-2)]"
        "a-D-Manp-(1-6)-[a-D-Manp-(1-3)]b-D-Manp-(1-4)-b-D-Glcp2NAc-(1-4)-b-D-Glcp2NAc")
    child = iupac.loads("b-D-Glcp2NAc")
    gntVI = Glycosyltransferase(4, 1, parent, child, terminal=False, parent_node_id=6,
                                identifying_information=enzdb[2, 4, 1, 145],
                                validator=rejecting(bisecting_glcnac, galactose))

    parent = iupac.loads("b-D-Glcp2NAc")
    child = iupac.loads("b-D-Galp")

    galt = Glycosyltransferase(4, 1, parent, child, identifying_information=enzdb[
                               2, 4, 1, 38], parent_node_id=1,
                               site_validator=reject_on_path(bisecting_glcnac))

    parent = iupac.loads("b-D-Galp-(1-4)-b-D-Glcp2NAc")
    child = iupac.loads("a-D-Galp")

    agal13galt = Glycosyltransferase(3, 1, parent, child, identifying_information=None,
                                     parent_node_id=3)

    parent = iupac.loads("b-D-Galp-(1-4)-b-D-Glcp2NAc")
    child = iupac.loads("b-D-Glcp2NAc")

    gntE = Glycosyltransferase(3, 1, parent, child, identifying_information=enzdb[
                               2, 4, 1, 149], parent_node_id=3)

    parent = iupac.loads("b-D-Galp-(1-4)-b-D-Glcp2NAc")
    child = iupac.loads("a-D-Neup5Ac")

    siat2_6 = Glycosyltransferase(6, 2, parent, child, identifying_information=enzdb[
                                  2, 4, 99, 1], parent_node_id=3)
    siat2_3 = Glycosyltransferase(3, 2, parent, child, identifying_information=enzdb[
                                  2, 4, 99, 6], parent_node_id=3)
    # siat2_8 = Glycosyltransferase(
    #     8, 2, child.clone(), child, identifying_information=enzdb[2, 4, 99, 8],
    #     site_validator=reject_on_parent(child.clone()), parent_node_id=1)

    # parent = iupac.loads("b-D-Glcp2NAc")
    # child = iupac.loads("b-D-Galp2NAc")
    # b4galnact = Glycosyltransferase(4, 1, parent, child, parent_node_id=1, identifying_information=None)

    parent = iupac.loads("b-D-Galp-(1-4)-b-D-Glcp2NAc")
    child = iupac.loads("a-L-Fucp")
    fuct3 = Glycosyltransferase(3, 1, parent, child, terminal=False,
                                parent_node_id=1, identifying_information=enzdb[2, 4, 1, 152])

    parent = iupac.loads(
        "beta-D-Glcp2NAc-(1->2)-alpha-D-Manp-(1->3)-[beta-D-Glcp2NAc-(1->2)-alpha-D-Manp-(1->6)]"
        "-beta-D-Manp-(1->4)-beta-D-Glcp2NAc-(1->4)-beta-D-Glcp2NAc")
    child = iupac.loads("a-L-Fucp")
    fuct6 = Glycosyltransferase(6, 1, parent, child, terminal=False,
                                parent_node_id=1, identifying_information=enzdb[2, 4, 1, 68])

    manI = Glycosylase((2,), 1, iupac.loads("a-D-Manp"), iupac.loads("a-D-Manp"),
                       identifying_information=enzdb["3.2.1.113"])
    manII = Glycosylase((3, 6), 1, iupac.loads("a-D-Manp"), iupac.loads("a-D-Manp"),
                        identifying_information=enzdb["3.2.1.114"],
                        validator=rejecting(bisecting_glcnac))
    glucosidaseI = Glycosylase(None, None, iupac.loads("alpha-D-Manp"), iupac.loads("?-?-Glcp"),
                               identifying_information=enzdb["3.2.1.106"])
    glucosidaseII = Glycosylase((3,), 1, iupac.loads("alpha-D-Glcp"), iupac.loads("alpha-D-Glcp"),
                                identifying_information=enzdb['3.2.1.84'])

    glycosylases = {
        "manI": manI,
        "manII": manII,
        "glucosidaseI": glucosidaseI,
        "glucosidaseII": glucosidaseII,
    }

    glycosyltransferases = {
        "gntI": gntI,
        "gntII": gntII,
        "gntIII": gntIII,
        "gntIV": gntIV,
        "gntV": gntV,
        "gntVI": gntVI,
        "galt": galt,
        "gntE": gntE,
        "siat2_6": siat2_6,
        "siat2_3": siat2_3,
        "fuct3": fuct3,
        "fuct6": fuct6,
        "agal13galt": agal13galt,
        # pending literature search: https://www.ncbi.nlm.nih.gov/pubmed/7881179
        # "b4galnact": b4galnact,
    }

    starting_structure_iupac = (
        "a-D-Manp-(1-2)-a-D-Manp-(1-6)-[a-D-Manp-(1-2)-a-D-Manp-(1-3)]"
        "a-D-Manp-(1-6)-[a-D-Glcp-(1-3)-a-D-Glcp-(1-3)-a-D-Glcp-(1-3)-"
        "a-D-Manp-(1-2)-a-D-Manp-(1-2)-a-D-Manp-(1-3)]"
        "b-D-Manp-(1-4)-b-D-Glcp2NAc-(1-4)-b-D-Glcp2NAc")
    starting_structure = iupac.loads(starting_structure_iupac)

    return glycosylases, glycosyltransferases, [starting_structure]


def make_mucin_type_o_glycan_pathway():
    enzdb = EnzymeDatabase._from_static()

    parent = iupac.loads("a-D-Galp2NAc")
    child = iupac.loads("b-D-Galp")
    c1galt1 = Glycosyltransferase(
        3, 1, parent, child, identifying_information=enzdb[2, 4, 1, 122])

    parent = iupac.loads("a-D-Galp2NAc")
    child = iupac.loads("b-D-Glcp2NAc")
    b3gnt6 = Glycosyltransferase(3, 1, parent, child, terminal=False,
                                 identifying_information=enzdb[2, 4, 1, 147])

    parent = iupac.loads("a-D-Galp2NAc")
    child = iupac.loads("b-D-Glcp2NAc")
    gcnt1 = Glycosyltransferase(6, 1, parent, child, terminal=False,
                                identifying_information=enzdb[2, 4, 1, 102])

    parent = iupac.loads('b-D-Glcp2NAc-(1-6)-a-D-Galp2NAc')
    child = iupac.loads("b-D-Galp")
    b4galt5 = Glycosyltransferase(4, 1, parent, child, parent_node_id=3)

    parent = iupac.loads('b-D-Glcp2NAc-(1-3)-a-D-Galp2NAc')
    child = iupac.loads('b-D-Glcp2NAc')
    gcnt3 = Glycosyltransferase(6, 1, parent, child, parent_node_id=1, terminal=False,
                                identifying_information=enzdb[2, 4, 1, 148])

    parent = iupac.loads("a-D-Galp2NAc")
    child = iupac.loads("a-D-Galp2NAc")
    core7_galnact = Glycosyltransferase(6, 1, parent, child)

    parent = iupac.loads("a-D-Galp2NAc")
    child = iupac.loads("a-D-Galp")
    core8_galt = Glycosyltransferase(3, 1, parent, child)

    parent = iupac.loads("a-D-Galp2NAc")
    child = iupac.loads("a-D-Neup5Ac")
    st6gal1 = Glycosyltransferase(6, 2, parent, child, terminal=False,
                                  identifying_information=enzdb[2, 4, 99, 3])

    parent = iupac.loads("b-D-Galp-(1-3)-a-D-Galp2NAc")
    child = iupac.loads("a-D-Neup5Ac")
    st3gal2 = Glycosyltransferase(3, 2, parent, child, parent_node_id=3)

    parent = iupac.loads("b-D-Galp-(1-3)-a-D-Galp2NAc")
    child = iupac.loads("a-D-Neup5Ac")
    st6gal2 = Glycosyltransferase(6, 2, parent, child, parent_node_id=3)

    parent = iupac.loads('b-D-Galp-(1-4)-b-D-Glcp2NAc')
    child = iupac.loads("a-L-Fucp")
    fuct2 = Glycosyltransferase(2, 1, parent, child, terminal=False, parent_node_id=3,
                                identifying_information=enzdb['2.4.1.69'])

    parent = iupac.loads("b-D-Galp-(1-4)-b-D-Glcp2NAc")
    child = iupac.loads("a-L-Fucp")
    fuct3 = Glycosyltransferase(3, 1, parent, child, terminal=False,
                                parent_node_id=1, identifying_information=enzdb[2, 4, 1, 152])

    parent = iupac.loads("b-D-Galp-(1-4)-b-D-Glcp2NAc")
    child = iupac.loads("a-D-Neup5Ac")
    siat2_6 = Glycosyltransferase(6, 2, parent, child, identifying_information=enzdb[
                                  2, 4, 99, 1], parent_node_id=3)

    parent = iupac.loads("b-D-Glcp2NAc")
    child = iupac.loads("b-D-Galp")
    galt4 = Glycosyltransferase(4, 1, parent, child, identifying_information=enzdb[
        2, 4, 1, 86])

    parent = iupac.loads("b-D-Galp-(1-4)-b-D-Glcp2NAc")
    child = iupac.loads("b-D-Glcp2NAc")
    gntE = Glycosyltransferase(3, 1, parent, child, identifying_information=enzdb[
        2, 4, 1, 149], parent_node_id=3)

    parent = iupac.loads("b-D-Galp-(1-4)-b-D-Glcp2NAc-(1-3)-b-D-Galp")
    child = iupac.loads("b-D-Glcp2NAc")
    gntII = Glycosyltransferase(6, 1, parent, child, terminal=False,
                                identifying_information=enzdb[2, 4, 1, 150], parent_node_id=1)

    parent = iupac.loads("b-D-Galp-(1-3)-a-D-Galp2NAc")
    child = iupac.loads("b-D-Glcp2NAc")
    b3gnt3 = Glycosyltransferase(3, 1, parent, child, identifying_information=enzdb[
        "2.4.1.102"], parent_node_id=3, terminal=False)

    glycosyltransferases = {
        "c1galt1": c1galt1,
        "b3gnt6": b3gnt6,
        "gcnt1": gcnt1,
        "b4galt5": b4galt5,
        "gcnt3": gcnt3,
        "core7_galnact": core7_galnact,
        "core8_galt": core8_galt,
        "st6gal1": st6gal1,
        "st3gal2": st3gal2,
        "st6gal2": st6gal2,
        "fuct2": fuct2,
        "fuct3": fuct3,
        "siat2_6": siat2_6,
        "galt4": galt4,
        "gntE": gntE,
        "gntII": gntII,
        "b3gnt3": b3gnt3
    }

    glycosylases = {}

    seeds = [glypy.Glycan(iupac.loads("a-D-Galp2NAc"))]
    return glycosylases, glycosyltransferases, seeds


ABS_Sialidase = Glycosylase((3, 6, 8, 9), 2, parent=None, child=iupac.loads("a-D-Neup5Ac"))
NAN1_Sialidase = Glycosylase((3,), 2, parent=None, child=iupac.loads("a-D-Neup5Ac"))
BKF_Fucosidase = Glycosylase((2, 3, 4, 6), 1, parent=None, child=iupac.loads("a-L-Fucp"))
XMF_Fucosidase = Glycosylase((2,), 1, parent=None, child=iupac.loads("a-L-Fucp"))
AMF_Fucosidase = Glycosylase((3, 4), 1, parent=None, child=iupac.loads("a-L-Fucp"))
BTG_Galactosidase = Glycosylase((3, 4), 1, parent=None, child=iupac.loads("b-D-Galp"))
SPG_Galactosidase = Glycosylase((4,), 1, parent=None, child=iupac.loads("b-D-Galp"))
CBG_Galactosidase = Glycosylase((3, 4, 6), 1, parent=None, child=iupac.loads("b-D-Galp"))
JBM_Mannosidase = Glycosylase((3, 4, 6), 1, parent=None, child=iupac.loads("a-D-Manp"))

GUH_N_Acetylhexosaminidase = Glycosylase(
    (2, 3, 4, 6), 1, parent=None, child=iupac.loads("b-D-GlcpNAc"),
    # Add site_validator-like option to Glycosylase
    # Block action on bisecting GlcNAc
    # site_validator=reject_on_path(
    #     iupac.loads("b-D-Glcp2NAc-(1-4)-b-D-Manp-(1-4)-b-D-Glcp2NAc-(1-4)-b-D-Glcp2NAc"))
)

JBH_N_Acetylhexosaminidase = Glycosylase((2, 3, 4, 6), 1, parent=None, child=(
    iupac.loads("b-D-GlcpNAc"),
    iupac.loads("b-D-GalpNAc")))

glycodigest_rules = {
    "ABS": ABS_Sialidase,
    "NAN1": NAN1_Sialidase,
    "BKF": BKF_Fucosidase,
    "XMF": XMF_Fucosidase,
    "AMF": AMF_Fucosidase,
    "BTG": BTG_Galactosidase,
    "SPG": SPG_Galactosidase,
    "CBG": CBG_Galactosidase,
    "JBM": JBM_Mannosidase,
    "GUH": GUH_N_Acetylhexosaminidase,
    "JBH": JBH_N_Acetylhexosaminidase
}



import unittest
import operator

import glypy
from glypy.composition import composition_transform
from glypy.algorithms import similarity

from .common import load


class SimilarityTests(unittest.TestCase):

    def test_deep_similarity(self):
        branchy = load("branchy_glycan")
        broad = load("broad_n_glycan")
        ref = broad.clone()
        self.assertEqual(similarity.monosaccharide_similarity(branchy.root, branchy.root), (7, 7))
        self.assertEqual(
            similarity.monosaccharide_similarity(branchy.root, branchy.root, include_children=True),
            (64, 64))
        self.assertEqual(similarity.monosaccharide_similarity(branchy.root, broad.root), (4, 7))
        self.assertEqual(
            similarity.monosaccharide_similarity(branchy.root, broad.root, include_children=True),
            (9, 14))
        self.assertEqual(
            similarity.monosaccharide_similarity(broad.root, branchy.root, include_children=True),
            (17, 20))
        self.assertEqual(similarity.monosaccharide_similarity(broad.root, broad.root, include_children=True), (91, 91))
        self.assertEqual(ref, broad)

    def test_build_index_pairs(self):
        nsc = similarity.NodeSimilarityComparator()
        pairs = {(18, 15): (9, 9), (15, 15): (9, 9), (15, 18): (9, 9), (18, 18): (9, 9)}
        expected = set([((18, 15), (15, 18)), ((18, 18), (15, 15))])
        result = nsc.build_unique_index_pairs(pairs)
        self.assertEqual(set(result), expected)

    def test_optimal_assignment(self):
        nsc = similarity.NodeSimilarityComparator()
        pairs = {(18, 15): (9, 9), (15, 15): (9, 9), (15, 18): (9, 9), (18, 18): (9, 9)}
        expected = set(((18, 15), (15, 18)))
        result = nsc.optimal_assignment(pairs)
        self.assertEqual(set(result), expected)

    def test_partial_similarity(self):
        broad = load("broad_n_glycan")
        expected = [
            (1, (91, 91)),
            (3, (84, 84)),
            (5, (77, 77)),
            (6, (39, 39)),
            (7, (20, 20)),
            (9, (6, 6)),
            (10, (7, 7)),
            (11, (13, 13)),
            (13, (6, 6)),
            (14, (32, 32)),
            (15, (13, 13)),
            (17, (6, 6)),
            (18, (13, 13)),
            (20, (6, 6))
        ]
        result = list(map(lambda x: (
            x[0].id, similarity.monosaccharide_similarity(
                x[0], x[1], include_children=1)), zip(broad, broad)))
        self.assertEqual(result, expected)

    def test_is_aminated(self):
        broad = load("broad_n_glycan")
        self.assertTrue(similarity.is_aminated(broad.root))
        self.assertTrue(similarity.has_substituent(broad.root, glypy.Substituent("n-acetyl")))
        self.assertFalse(similarity.is_aminated(broad[3]))

    def test_is_reduced(self):
        broad = load("broad_n_glycan")
        broad.reducing_end = True
        self.assertTrue(similarity.is_reduced(broad))
        self.assertFalse(similarity.is_reduced(None))

    def test_ignore_reduction(self):
        broad = load("broad_n_glycan")
        broad.reducing_end = True
        self.assertTrue(similarity.is_reduced(broad))
        r, q = similarity.monosaccharide_similarity(
            broad.root, broad.root)
        r2, q2 = similarity.monosaccharide_similarity(
            broad.root, broad.root, ignore_reduction=True)
        self.assertTrue((r - 1) == r2)

    def test_has_monosaccharide(self):
        broad = load("broad_n_glycan")
        self.assertTrue(similarity.has_fucose(broad))
        self.assertTrue(similarity.has_modification(broad[6], 'd'))
        self.assertFalse(similarity.has_modification(broad[5], 'd'))

    def test_similar_substituents(self):
        self.assertTrue(
            similarity.monosaccharide_similarity(
                glypy.Substituent('n-acetyl'),
                glypy.Substituent('n-acetyl')) == (1, 1))

    def test_is_derivatized(self):
        broad = load("broad_n_glycan")
        self.assertFalse(similarity.is_derivatized(broad.root))
        composition_transform.derivatize(broad, 'methyl')
        self.assertTrue(similarity.is_derivatized(broad.root))


if __name__ == '__main__':
    unittest.main()



import itertools
import re

from glypy.composition import Composition

_fragment_shift = {
    "B": Composition(O=1, H=2),
    "Y": Composition(),
    "C": Composition(),
    "Z": Composition(H=2, O=1),
}


def _link_ids_splitter(fragment, link_ids, kind):  # pragma: no cover
    ion_types = re.findall(r"(\d+,\d+)?(\S)", kind)
    links_broken = link_ids

    pairings = zip(ion_types, links_broken)

    fragment.link_ids = {
        link_id: ion_type for ion_type,
        link_id in pairings if ion_type[0] == ""}
    fragment.crossring_cleavages = {
        node_id: ion_type for ion_type,
        node_id in pairings if ion_type[0] != ""}


latex_symbol_map = {
    "a": r"\alpha",
    "b": r"\beta",
    "c": r"\sigma",
    "d": r"\delta",
    "e": r"\epsilon",
    "f": r"\zeta",
    "g": r"\gamma",
    "h": r"\eta",
    "i": r"\iota",
    "j": r"\lambda",
    "k": r"\kappa",
    "l": r"\lambda",
    "m": r"\mu",
    "n": r"\nu",
    "o": r"\omicron",
    "p": r"\pi",
    "q": r"\sigma",
    "r": r"\rho",
    "s": r"\upsilon",
    "t": r"\tau",
    "u": r"\upsilon",
    "v": r"\omega",
}


class GlycanFragment(object):
    '''
    A simple container for a fragment ion, produced by :meth:`Glycan.fragments`

    Attributes
    ----------
    kind: |str|
        One of A, B, C, X, Y, or Z for each link broken or ring cleaved

    link_ids: |dict| of |int| -> |tuple|
        The :attr:`id` value of each link cleaved to the corresponding cleavage type

    included_nodes: |list| of |int|
        The :attr:`id` value of each |Monosaccharide| contained in the fragment

    mass: |float|
        The mass of the fragment

    name: |str|
        The fragment name under the branching nomenclature

    crossring_cleavages: |dict| of |int| -> |tuple|
        The :attr:`id` value of each link cleaved to the corresponding cleavage type, including ring coordinates

    composition: |Composition|
        The elemental composition of the fragment

    See Also
    --------
    :meth:`Glycan.fragments`
    '''
    __slots__ = [
        "mass",
        "kind",
        "included_nodes",
        "link_ids",
        "name",
        "crossring_cleavages",
        "composition",
    ]

    def __init__(self, kind, link_ids, included_nodes, mass,
                 name=None, crossring_cleavages=None, composition=None):
        self.mass = mass
        self.kind = kind
        self.link_ids = link_ids
        self.included_nodes = included_nodes
        self.crossring_cleavages = crossring_cleavages
        self.name = name
        self.composition = composition

    def copy(self):
        """Create a deep copy of the fragment.

        Returns
        -------
        GlycanFragment
        """
        return self.__class__(
            self.kind, self.link_ids.copy(), self.included_nodes.copy(), self.mass,
            self.name, self.crossring_cleavages.copy(), self.composition.copy())

    def clone(self):
        """Create a deep copy of the fragment.

        Returns
        -------
        GlycanFragment
        """
        return self.copy()

    def is_reducing(self):
        """Is this fragment from the reducing end

        Returns
        -------
        |bool|
        """
        return bool(set(self.kind) & set("XYZ"))

    def is_non_reducing(self):
        """Is this fragment from the non-reducing end

        Returns
        -------
        |bool|
        """
        return bool(set(self.kind) & set("ABC"))

    def is_internal(self):
        """Is the fragment composed of both reducing and non-reducing bond
        cleavage events.

        Returns
        -------
        bool
        """
        return bool(self.is_reducing() and self.is_non_reducing())

    @property
    def fname(self):
        """A `LaTeX` formatted version of the fragment's name

        Returns
        -------
        str
        """
        buff = []
        for c in self.name:
            if c in latex_symbol_map:
                buff.append("$_{}$".format(latex_symbol_map[c]))
            else:
                buff.append(c)
        return ''.join(buff)

    def _make_tex_name(self):
        buff = []
        for c in self.name:
            if c in latex_symbol_map:
                buff.append("_{}".format(latex_symbol_map[c]))
            else:
                buff.append(c)
        return ''.join(buff)

    @property
    def series(self):
        """An alias for :attr:`kind`

        Returns
        -------
        str
        """
        return self.kind

    def __getstate__(self):  # pragma: no cover
        d = {}
        for a in self.__slots__:
            d[a] = getattr(self, a)
        return d

    def __setstate__(self, state):  # pragma: no cover
        for a, v in state.items():
            setattr(self, a, v)

    def __hash__(self):
        return hash((self.name, int(self.mass)))

    def __eq__(self, other):  # pragma: no cover
        for field in self.__slots__:
            if getattr(self, field) != getattr(other, field, NotImplemented):
                return False
        return True

    def __ne__(self, other):  # pragma: no cover
        return not self == other

    def __repr__(self):  # pragma: no cover
        rep = "<GlycanFragment "
        for f in self.__slots__:
            rep += " {}={}".format(f, getattr(self, f))
        rep += ">"
        return rep

    @property
    def __dict__(self):
        return self.__getstate__()

    @property
    def break_count(self):
        """The number of cleavage events to create this fragment.

        Returns
        -------
        int
        """
        return len(self.link_ids) + len(self.crossring_cleavages)

    @property
    def residues_contained(self):
        """The number of :class:`~.Monosaccharide` nodes included in the fragmented sub-tree.

        Returns
        -------
        int
        """
        return len(self.included_nodes)


Fragment = GlycanFragment


class GlycanSubstructure(object):
    """Represent a subgraph of a :class:`~.Glycan` produced
    from another :class:`~.Glycan` by breaking linkages.

    Supports the :func:`~.root` and :func:`~.tree` protocols.

    Attributes
    ----------
    tree : :class:`~.Glycan`
        The substructure
    include_nodes : set
        The :attr:`~.Monosaccharide.id` of all included monosaccharides
    link_ids : list
        The :attr:`~.Link.id` of all linkages broken to form this substructure
    parent_breaks : dict
        Mapping from :attr:`~.Link.id` to :attr:`~.Monosaccharide.id` if that monosaccharide
        was the :attr:`~.Link.parent` of that linkage and that linkage was broken.
    child_breaks : dict
        Mapping from :attr:`~.Link.id` to :attr:`~.Monosaccharide.id` if that monosaccharide
        was the :attr:`~.Link.child` of that linkage and that linkage was broken.
    crossring_cleavages : dict
        Mapping from :attr:`~.Monosaccharide.id` to :class:`str` denoting the ring coordinate pair that
        was cleaved to produce any included crossring cleavage events.

    """

    def __init__(self, tree, include_nodes, link_ids,
                 parent_breaks, child_breaks, crossring_cleavages=None):
        self.tree = tree
        self.include_nodes = include_nodes
        self.link_ids = link_ids
        self.parent_breaks = parent_breaks
        self.child_breaks = child_breaks
        self.crossring_cleavages = crossring_cleavages or {}

    def contains_reducing_end(self):
        """Whether or not the sub-tree includes a :class:`ReducedEnd`

        Returns
        -------
        bool
        """
        return self.tree.root.reducing_end is not None

    def to_fragments(self, kind="BY", average=False, charge=None, mass_data=None, include_composition=True,
                     traversal_method='dfs'):
        """Transform an instance of :class:`GlycanSubstructure` into every combination of
        :class:`GlycanFragment` allowed under `kind`.

        Parameters
        ----------
        kind : Iterable, optional
            The types of fragments to emit. Defaults to "BY"
        average : bool, optional
            Calculate masses with average isotopic composition
        charge : int, optional
            Calculate `m/z` instead of neutral mass, with `z = charge`
        mass_data : dict, optional
            If mass_data is None, standard NIST mass and isotopic abundance data are used. Otherwise the
            contents of mass_data are assumed to contain elemental mass and isotopic abundance information.
            Defaults to :const:`None`.
        include_composition: bool, optional
            Whether or not to populate the `composition` attribute of the fragment. Defaults to :const:`True`

        Yields
        -------
        :class:`GlycanFragment`
        """
        parent_type = set("YZ") & set(kind)
        child_type = set("BC") & set(kind)
        crossring_type = set("AX") & set(kind)
        parent_shifts = [(lid, parent_type)
                         for lid, nid in self.parent_breaks.items()]
        child_shifts = [(lid, child_type)
                        for lid, nid in self.child_breaks.items()]

        shift_masses = {k: _fragment_shift[k].calc_mass(average=average,
                                                        charge=charge,
                                                        mass_data=mass_data)
                        for k in parent_type | child_type}
        crossring_contained = {
            kind for pos,
            kind in self.crossring_cleavages.values()}
        # Do not proceed if crossring fragments are included and
        # were not requested.
        if len(crossring_contained & crossring_type) == 0 and len(
                crossring_contained) != 0:
            # raise StopIteration()
            return

        all_shifts = parent_shifts + child_shifts
        all_link_ids = [i for i, t in all_shifts]
        frag_types = [t for i, t in all_shifts]
        base_mass = self.tree.mass(
            average=average,
            charge=charge,
            mass_data=mass_data, method=traversal_method)
        if include_composition:
            base_composition = self.tree.total_composition(method=traversal_method)
        # product of splat of empty list is a list of the empty list. So a fragment with
        # no glycosidic cleavages still enters this outer loop, letting only crossring-cleavage
        # GlycanSubstructure instances through without issue
        for shift_set in itertools.product(*frag_types):
            mass_offset = 0.0
            composition_offset = Composition()
            link_ids = {}
            # The type of fragment being produced, expressed a collection of ABCXYZs
            kind = [] + [''.join(cr_kind)
                         for cr_kind in self.crossring_cleavages.values()]
            i = 0
            shift_set = list(shift_set)
            for shift in shift_set:
                link_id = all_link_ids[i]
                shift = shift[0]
                mass_offset -= shift_masses[shift]
                composition_offset -= _fragment_shift[shift]
                link_ids[link_id] = ("", shift)
                kind.append(shift)
                i += 1

            if include_composition:
                fragment_composition = base_composition + composition_offset

            yield GlycanFragment(kind=''.join(kind), link_ids=link_ids, included_nodes=self.include_nodes,
                                 mass=base_mass + mass_offset, name=None,
                                 crossring_cleavages=self.crossring_cleavages,
                                 composition=fragment_composition)

    def __repr__(self):  # pragma: no cover
        rep = ("<GlycanSubstructure include_nodes={} link_ids={} parent_breaks={}"
               " child_breaks={} crossring_cleavages={}>\n{}").format(
                   self.include_nodes, self.link_ids, self.parent_breaks,
                   self.child_breaks, self.crossring_cleavages, self.tree)
        return rep

    def __root__(self):  # pragma: no cover
        return self.tree.root

    def __tree__(self):  # pragma: no cover
        return self.tree


Subtree = GlycanSubstructure



from glypy._c.structure.base cimport SaccharideBase
from glypy.composition.ccomposition cimport CComposition
from glypy.utils.cenum cimport EnumValue


cdef class Monosaccharide(SaccharideBase):
    cdef:
        public long id
        public EnumValue _anomer
        public EnumValue _superclass
        public int ring_start
        public int ring_end
        public object links
        public object substituent_links
        public object modifications
        public CComposition composition
        public object _reducing_end
        public object _degree
        public bint _checked_for_reduction


